IMPLEMENTATION MODULE TrackDiskSupport;

FROM SYSTEM      IMPORT ADR, LONGSET, ADDRESS, BITSET;

FROM Arts        IMPORT Assert, TermProcedure, detectCtrlC;

FROM Exec        IMPORT DoIO, OpenDevice, CloseDevice, MsgPortPtr, Byte;

FROM ExecSupport IMPORT BeginIO, AbortIO, CreatePort, DeletePort,
                        CreateExtIO, DeleteExtIO;

FROM TrackDisk   IMPORT motor, seek, remove, changeNum, changeState,
                        extClear, protStatus, rawRead, rawWrite,
                        getDriveType, getNumTracks, addChangeInt,
                        remChangeInt, lastComm, IOExtTDPtr, sector,
                        extRead, extWrite, extUpdate, numSecs, extFormat,
                        allowNon35;

CONST
  tdname = "trackdisk.device";

VAR
  DiskPort: ARRAY[0..3] OF MsgPortPtr;
  DiskReqBlock: ARRAY[0..3] OF IOExtTDPtr;
  dev: CARDINAL;
  OpenDrives: BITSET;
  LastDCC: BOOLEAN;
  CtrlCCnt: CARDINAL;

(*------  CtrlC-Control:  ------*)

PROCEDURE ForbidCC();

BEGIN
  IF CtrlCCnt=0 THEN
    LastDCC := detectCtrlC;
    detectCtrlC := FALSE;
  END;
  INC(CtrlCCnt);
END ForbidCC;

PROCEDURE PermitCC();

BEGIN
  DEC(CtrlCCnt);
  IF CtrlCCnt=0 THEN
    detectCtrlC := LastDCC;
  END;
END PermitCC;

(*------  Open TrackDisk:  ------*)

PROCEDURE OpenTrackDisk(device: INTEGER; AllowNon35: BOOLEAN): Byte;

(* OpenTrackDisk can be called up to 4 times with 4 different device-      *)
(* numbers (0 to 3). Higher values are not allowed.                        *)
(* AllowNon35 specifies whether 5.25 Drives should be opened or not        *)
(**** How can a drive be allocated and removed from Devicelist etc?  *******)

VAR
  err: Byte;

BEGIN
  ForbidCC();
  Assert(NOT(device IN OpenDrives),ADR("Can't open a Drive twice !!!"));
  DiskPort[device] := CreatePort(NIL,0);
  Assert(DiskPort[device]#NIL,ADR("CreatePort failed"));
  DiskReqBlock[device] := CreateExtIO(DiskPort[device],SIZE(DiskReqBlock[device]^));
  Assert(DiskReqBlock[device]#NIL,ADR("CreateExtIO failed"));
  IF AllowNon35 THEN
    OpenDevice(ADR(tdname),device,DiskReqBlock[device],LONGSET{allowNon35});
  ELSE
    OpenDevice(ADR(tdname),device,DiskReqBlock[device],LONGSET{});
  END;
  err := DiskReqBlock[device]^.req.error;
  IF err#0 THEN
    DeleteExtIO(DiskReqBlock[device]);
    DiskReqBlock[device] := NIL;
    DeletePort(DiskPort[device]);
    DiskPort[device] := NIL;
  ELSE
    INCL(OpenDrives,device);
  END;
  PermitCC();
  RETURN err;
END OpenTrackDisk;

(*------  Close TrackDisk:  ------*)

PROCEDURE CloseTrackDisk(device: INTEGER);

BEGIN
  ForbidCC();
  Assert(device IN OpenDrives,ADR("Can't close unopened drive !!!"));
  OpenDrives := OpenDrives - {device};
  CloseDevice(DiskReqBlock[device]);
  DeleteExtIO(DiskReqBlock[device]);
  DiskReqBlock[device] := NIL;
  DeletePort(DiskPort[device]);
  DiskPort[device] := NIL;
  PermitCC();
END CloseTrackDisk;

(*------  Get Disk Change Value:  ------*)

PROCEDURE GetDiskChange(device: INTEGER): LONGCARD;

BEGIN
  ForbidCC();
  DiskReqBlock[device]^.req.command := changeNum;
  DoIO(DiskReqBlock[device]);
  PermitCC();
  RETURN DiskReqBlock[device]^.req.actual;
END GetDiskChange;

(*------  Is there a disk inserted?  ------*)

PROCEDURE ChangeState(device: INTEGER): BOOLEAN;
(* returns TRUE if disk in drive *)

BEGIN
  ForbidCC();
  DiskReqBlock[device]^.req.command := changeState;
  DoIO(DiskReqBlock[device]);
  PermitCC();
  RETURN 0=DiskReqBlock[device]^.req.actual;
END ChangeState;

(*------  Test Write-Protection:  ------*)

PROCEDURE ProtStatus(device: INTEGER): BOOLEAN;
(* TRUE if disk is Write enabled *)

BEGIN
  ForbidCC();
  DiskReqBlock[device]^.req.command := protStatus;
  DoIO(DiskReqBlock[device]);
  PermitCC();
  RETURN 0=DiskReqBlock[device]^.req.actual;
END ProtStatus;

(*------  Motor Control:  ------*)

PROCEDURE Motor(device: INTEGER; On: BOOLEAN): BOOLEAN;
(* Turns Diskmotor on or off according to On. Result is old motor state:   *)
(* TRUE = on, FALSE = off. *)

BEGIN
  ForbidCC();
  IF On THEN
    DiskReqBlock[device]^.req.length := 1;
  ELSE
    DiskReqBlock[device]^.req.length := 0;
  END;
  DiskReqBlock[device]^.req.command := motor;
  DoIO(DiskReqBlock[device]);
  PermitCC();
  RETURN DiskReqBlock[device]^.req.actual=1;
END Motor;

(*------  Seek:  ------*)

PROCEDURE Seek(device: INTEGER; Offset: LONGCARD): Byte;
(* Return error code *)

BEGIN
  ForbidCC();
  DiskReqBlock[device]^.req.offset := Offset;
  DiskReqBlock[device]^.req.command:= seek;
  DoIO(DiskReqBlock[device]);
  PermitCC();
  RETURN DiskReqBlock[device]^.req.error;
END Seek;

(*------  Read Block:  ------*)

PROCEDURE ReadBlock(device: INTEGER;
                    Block: INTEGER;
                    Buffer: ADDRESS;
                    BlockCnt: CARDINAL;
                    ChangeCnt: LONGCARD): Byte;

(* Block:        the disk's block # to be read                             *)
(* Buffer:       at least 512 Bytes of Chip memory to store data           *)
(* BlockCnt:     Number of Blocks to be read                               *)
(* ChangeCnt:    last changenumber to check diskchanges                    *)
(* result:       error number                                              *)


BEGIN
  ForbidCC();
  WITH DiskReqBlock[device]^ DO
    req.length := LONGCARD(BlockCnt)*sector;
    req.data   := Buffer;
    req.command:= extRead;
    count      := ChangeCnt;
    req.offset := sector*LONGCARD(Block);
  END;
  DoIO(DiskReqBlock[device]);
  PermitCC();
  RETURN DiskReqBlock[device]^.req.error;
END ReadBlock;

(*------  Read Sector:  ------*)

PROCEDURE ReadCycSec(device: INTEGER;
                     cyl,sec,head: INTEGER;
                     Buffer: ADDRESS;
                     BlockCnt: CARDINAL;
                     ChangeCnt: LONGCARD): Byte;

(* cyl,sec,head: specify the block to be read                              *)
(* Buffer:       512 Bytes of Chip memory to store data                    *)
(* BlockCnt:     Number of Blocks to be read                               *)
(* ChangeCnt:    last changenumber to check diskchanges                    *)
(* result:       error number                                              *)


BEGIN
  RETURN ReadBlock(device,sec+numSecs*(head+2*cyl),Buffer,BlockCnt,ChangeCnt);
END ReadCycSec;

(*------  Write Block:  ------*)

PROCEDURE WriteBlock(device: INTEGER;
                     block: INTEGER;
                     Buffer: ADDRESS;
                     BlockCnt: CARDINAL;
                     ChangeCnt: LONGCARD): Byte;

(* block:        the block # to be written                                 *)
(* Buffer:       512 Bytes of Chip memory storing the data                 *)
(* BlockCnt:     # of blocks to be written                                 *)
(* ChangeCnt:    last changenumber to check diskchanges                    *)
(* result:       error number                                              *)

BEGIN
  ForbidCC();
  WITH DiskReqBlock[device]^ DO
    req.length := LONGCARD(BlockCnt)*sector;
    req.data   := Buffer;
    req.command:= extWrite;
    count      := ChangeCnt;
    req.offset := sector*LONGCARD(block);
  END;
  DoIO(DiskReqBlock[device]);
  PermitCC();
  RETURN DiskReqBlock[device]^.req.error;
END WriteBlock;

(*------  Write Sector:  ------*)

PROCEDURE WriteCycSec(device: INTEGER;
                      cyl,sec,head: INTEGER;
                      Buffer: ADDRESS;
                      BlockCnt: CARDINAL;
                      ChangeCnt: LONGCARD): Byte;

(* cyl,sec,head: specify the block to be written                           *)
(* Buffer:       512 Bytes of Chip memory storing the data                 *)
(* BlockCnt:     # of blocks to be written                                 *)
(* ChangeCnt:    last changenumber to check diskchanges                    *)
(* result:       error number                                              *)


BEGIN
  RETURN WriteBlock(device,sec+numSecs*(head+2*cyl),Buffer,BlockCnt,ChangeCnt);
END WriteCycSec;

(*------  Update:  ------*)

PROCEDURE Update(device: INTEGER;
                 ChangeCnt: LONGCARD): Byte;

BEGIN
  ForbidCC();
  WITH DiskReqBlock[device]^ DO
    req.command:= extUpdate;
    count      := ChangeCnt;
  END;
  DoIO(DiskReqBlock[device]);
  PermitCC();
  RETURN DiskReqBlock[device]^.req.error;
END Update;

(*------  Format:  ------*)

PROCEDURE FormatTrack(device: INTEGER;
                 cyl: INTEGER;
                 Buffer: ADDRESS;
                 CycCnt: CARDINAL;
                 ChangeCnt: LONGCARD): Byte;

(* cyl:       First Cyclinder to be formatted                              *)
(* Buffer:    Buffer containing cyclinder data (at least 22*512 bytes)     *)
(* CycCnt:    Number of Tracks to be formatted                             *)
(* ChangeCnt: last changenumber to check diskchanges                       *)
(* result:    error number                                                 *)


BEGIN
  ForbidCC();
  WITH DiskReqBlock[device]^ DO
    req.length := LONGCARD(CycCnt)*sector*numSecs*2;
    req.data   := Buffer;
    req.command:= extFormat;
    count      := ChangeCnt;
    req.offset := sector*numSecs*2*LONGINT(cyl);
  END;
  DoIO(DiskReqBlock[device]);
  PermitCC();
  RETURN DiskReqBlock[device]^.req.error;
END FormatTrack;

(*------  Create Diskremove-Interrupt:  ------*)

PROCEDURE Remove(device: INTEGER;
                 IntProc: ADDRESS): Byte;

(* IntProc: The Interrupt Procedure's address or NIL to remove it          *)

BEGIN
  ForbidCC();
  WITH DiskReqBlock[device]^ DO
    req.data   := IntProc;
    req.command:= remove;
  END;
  DoIO(DiskReqBlock[device]);
  PermitCC();
  RETURN DiskReqBlock[device]^.req.error;
END Remove;

(*------  Mark Track as invalid:  ------*)

PROCEDURE Clear(device: INTEGER): Byte;

BEGIN
  ForbidCC();
  DiskReqBlock[device]^.req.command := extClear;
  DoIO(DiskReqBlock[device]);
  PermitCC();
  RETURN DiskReqBlock[device]^.req.error;
END Clear;

(*------  3.5 or 5.25 Drive:  ------*)

PROCEDURE GetDriveType(device: INTEGER): LONGCARD;

(* Returns 0 if 3.5 inch drive connected, else 1 (5.25")                   *)

BEGIN
  ForbidCC();
  DiskReqBlock[device]^.req.command := getDriveType;
  DoIO(DiskReqBlock[device]);
  PermitCC();
  RETURN DiskReqBlock[device]^.req.actual;
END GetDriveType;

(*------  How many Trax:  ------*)

PROCEDURE GetNumTracks(device: INTEGER): LONGCARD;

BEGIN
  ForbidCC();
  DiskReqBlock[device]^.req.command := getNumTracks;
  DoIO(DiskReqBlock[device]);
  PermitCC();
  RETURN DiskReqBlock[device]^.req.actual;
END GetNumTracks;

(*------  CleanUp:  ------*)

PROCEDURE CleanUp();

BEGIN
  FOR dev:=0 TO 3 DO
    IF dev IN OpenDrives     THEN CloseTrackDisk(dev) END;
    IF DiskReqBlock[dev]#NIL THEN DeleteExtIO(DiskReqBlock[dev]) END;
    IF DiskPort[dev]    #NIL THEN DeletePort(DiskPort[dev]) END;
  END;
END CleanUp;

(*------  Initialization:  ------*)

BEGIN
  FOR dev := 0 TO 3 DO
    DiskReqBlock[dev] := NIL;
    DiskPort[dev] := NIL;
  END;
  CtrlCCnt := 0;
  OpenDrives := {};
  TermProcedure(CleanUp);
END TrackDiskSupport.
