(**********************************************************************

    :Program.    Interface.mod
    :Contents.   Interface between m2make, m2c and m2l
    :Support.    thanks to Fridtjof for his contribution to "Startm2c"
    :Author.     Nicolas Benezan [bne]
    :Address.    Postwiesenstr. 2, D7000 Stuttgart 60
    :Phone.      711/333679
    :Copyright.  Shareware, see Doc-File
    :Language.   Modula-2
    :Translator. M2Amiga A+L V3.2d
    :Imports.    Environment, MemSystem1.3, ErrorReq1.2 [bne]
    :Imports.    Queue1.1 [mif]
    :History.    V1.0c [bne] 11.Feb.1989
    :Update.     22.Feb.1989 [bne] cosmetics
    :Update.     4.Mar.1989 [bne] adapted MemSystem V1.3
    :History.    V1.1a [bne] 30.Mar.1989 (CreateProc bug fixed)
    :History.    V1.2b [bne] 5.Apr.1989 (+ Options, cosmetics)

**********************************************************************)

IMPLEMENTATION MODULE Interface;

FROM SYSTEM      IMPORT ADR, ADDRESS, BPTR;
FROM Arts        IMPORT ModName, TermProcedure;
FROM Dos         IMPORT LoadSeg, UnLoadSeg, CreateProc, sharedLock,
                 Lock, UnLock, DupLock, Process, FileLockPtr, CurrentDir;
FROM Environment IMPORT SystemDir, OptionFlagSet, OptionFlags,
                 GetArguments, CompilerFlags, LinkerFlags, OptionChars;
FROM ErrorReq    IMPORT Assert;
FROM Exec        IMPORT Message, MessagePtr, MsgPort, MsgPortPtr, PutMsg,
                        WaitPort, GetMsg, Task, NodeType;
FROM ExecSupport IMPORT CreatePort, DeletePort;
FROM MemSystem   IMPORT NoCareAllocate, Deallocate;
FROM Queue       IMPORT queue,Init,Discard,Read,Write;
FROM Strings     IMPORT Copy;
FROM Workbench   IMPORT WBStartup, WBStartupPtr, WBArg, WBArgumentsPtr;

TYPE
  ProcessPtr = POINTER TO Process;
  CompArgument = RECORD
    name:ModName;
    dir:FileLockPtr;
  END;

CONST
  CompilerStack=30000;
  LinkerStack=15000;
  Priority=0;

  (* Error messages *)
  QueueInitFail   = "InitCompilerArgList() failed";
  QueueWriteFail  = "AddCompilerArg() failed";
  LoadPrgFail     = "could not load m2c/m2l";
  UnexpectedMsg   = "unexpected Message received";
  CreatePortFail  = "CreatePort() failed";

VAR
  MsgOutstanding: BOOLEAN;
  Segment: BPTR;
  procPort: MsgPortPtr;
  MyPort: MsgPortPtr;
  proc: ProcessPtr;
  StartupMsg: WBStartup;
  CompArgList:queue;     (* FIFO list *)
  NumCompArgs:INTEGER;
  NameList:POINTER TO ARRAY [1..255] OF ModName;

(**)
(* load code and create process                                         *)
(**)
PROCEDURE StartProcess(Name:ADDRESS;
                       Arguments:WBArgumentsPtr;
                       NumArgs:INTEGER;
                       StackSize:LONGINT);
  BEGIN
    Segment:=LoadSeg(Name);
    Assert(Segment#NIL,ADR(LoadPrgFail));
    procPort:=CreateProc(Name,Priority,Segment,StackSize);
    Assert(procPort#NIL,ADR(LoadPrgFail));
    proc := ADDRESS(procPort); (* address of the MsgPort of process *)
    DEC(proc,SIZE(Task));      (* address of process *)
    WITH StartupMsg.message DO
      node.type := message;
      replyPort := MyPort;
      length := SIZE(WBStartup);
    END;
    WITH StartupMsg DO
      process := procPort;
      segment := Segment;
      numArgs := NumArgs;
      toolWindow := NIL;
      argList := Arguments;
    END;
    MsgOutstanding:=TRUE;
    PutMsg(procPort,ADR(StartupMsg));
  END StartProcess;

PROCEDURE MakeOptionString(Flags,Mask:OptionFlagSet;
                           VAR String:ARRAY OF CHAR):BOOLEAN;
  VAR
    OptChar:ARRAY [0..0] OF CHAR;
    Flag:OptionFlags;
    Pos:INTEGER;
  BEGIN
    IF Flags*Mask#OptionFlagSet{} THEN
      String[0]:="-";
      Pos:=1;
      FOR Flag:=MIN(OptionFlags) TO MAX(OptionFlags) DO
        IF Flag IN Flags*Mask THEN
          Copy(OptChar,OptionChars,ORD(Flag),1);
          String[Pos]:=OptChar[0];
          INC(Pos);
        END;
      END;
      String[Pos]:=CHR(0);
      RETURN TRUE;
    ELSE
      RETURN FALSE;
    END;
  END MakeOptionString;

PROCEDURE InitCompilerArgList;
  VAR
    OptStr:ModName;
    Options:OptionFlagSet;
  BEGIN
    Assert(Init(CompArgList),ADR(QueueInitFail));
    NumCompArgs:=0;
    GetArguments(OptStr,Options);
    IF MakeOptionString(Options,CompilerFlags,OptStr) THEN
      AddCompilerArg(OptStr,NIL);
    END;
  END InitCompilerArgList;

PROCEDURE AddCompilerArg(Argument:ModName;Dir:FileLockPtr);
  VAR
    NewArg:CompArgument;
  BEGIN
    WITH NewArg DO
      name:=Argument;
      IF Dir#NIL THEN
        dir:=DupLock(Dir);
      ELSE
        dir:=NIL;
      END;
    END;
    IF NOT Write(CompArgList,NewArg) THEN
      UnLock(NewArg.dir);
      Assert(FALSE,ADR(QueueWriteFail));
    END;
    INC(NumCompArgs);
  END AddCompilerArg;

PROCEDURE StartCompiler;
  VAR
    oldDir:FileLockPtr;
    ArgList:WBArgumentsPtr;
    ArgNum:INTEGER;
    Arg:CompArgument;
  BEGIN
    NoCareAllocate(ArgList,(NumCompArgs+1)*SIZE(WBArg));
    NoCareAllocate(NameList,NumCompArgs*SIZE(ModName));
    WITH ArgList^[0] DO
      lock:=SystemDir;
      name:=ADR(CompilerName);
    END;
    FOR ArgNum:=1 TO NumCompArgs DO
      Read(CompArgList,Arg);
      NameList^[ArgNum]:=Arg.name;
      WITH ArgList^[ArgNum] DO
        lock:=Arg.dir;
        name:=ADR(NameList^[ArgNum]);
      END;
    END;
    Discard(CompArgList);
    oldDir:=CurrentDir(SystemDir);
    StartProcess(ADR(CompilerName),ArgList,NumCompArgs+1,CompilerStack);
  END StartCompiler;

PROCEDURE StartLinker(Argument:ModName;
                      Dir:FileLockPtr);
  VAR
    oldDir:FileLockPtr;
    ArgList:WBArgumentsPtr;
    Dummy:ModName;
    Options:OptionFlagSet;
    ArgNum:INTEGER;
  BEGIN
    NoCareAllocate(ArgList,SIZE(WBArg)*3);
    NoCareAllocate(NameList,SIZE(ModName)*2);
    WITH ArgList^[0] DO
      lock:=SystemDir;
      name:=ADR(LinkerName);
    END;
    GetArguments(Dummy,Options);
    IF MakeOptionString(Options,LinkerFlags,NameList^[1]) THEN
      WITH ArgList^[1] DO
        lock:=NIL;
        name:=ADR(NameList^[1]);
      END;
      ArgNum:=2;
    ELSE
      ArgNum:=1;
    END;
    NameList^[ArgNum]:=Argument;
    WITH ArgList^[ArgNum] DO
      lock:=DupLock(Dir);
      name:=ADR(NameList^[ArgNum]);
    END;
    oldDir:=CurrentDir(SystemDir);
    StartProcess(ADR(LinkerName),ArgList,ArgNum+1,LinkerStack);
(*    oldDir:=CurrentDir(oldDir); *)
  END StartLinker;

PROCEDURE WaitUntilCompleted;
  VAR
    ArgNum:INTEGER;
  BEGIN
    WaitPort(MyPort);
    Assert(GetMsg(MyPort)=ADR(StartupMsg),ADR(UnexpectedMsg));
    MsgOutstanding:=FALSE;
    WITH StartupMsg DO
      FOR ArgNum:=1 TO numArgs-1 DO
        IF argList^[ArgNum].lock#NIL THEN
          UnLock(argList^[ArgNum].lock);
        END;
      END;
      Deallocate(argList);
    END;
    IF NameList#NIL THEN
      Deallocate(NameList);
    END;
    UnLoadSeg(Segment);
  END WaitUntilCompleted;

(**)
(* termination procedure                                                *)
(**)
PROCEDURE CleanUp;
  BEGIN
    IF MsgOutstanding THEN
      WaitUntilCompleted;
    END;
    DeletePort(MyPort);
  END CleanUp;

BEGIN
  MyPort:=CreatePort(NIL,0);
  Assert(MyPort#NIL,ADR(CreatePortFail));
  MsgOutstanding:=FALSE;
  NameList:=NIL;
  TermProcedure(CleanUp);
END Interface.

