(**********************************************************************

    :Program.    Make.mod
    :Contents.   m2make - automagical compiler + linker
    :Support.    Michael Frie [mif]
    :Author.     Nicolas Benezan [bne]
    :Address.    Postwiesenstr. 2, D7000 Stuttgart 60
    :Phone.      711/333679
    :Copyright.  Shareware, see Doc-File for details
    :Language.   Modula-2
    :Translator. M2Amiga AMSoft V3.2d
    :Imports.    Queue1.2, AVL1.2 [mif]
    :Imports.    Environment1.2, PathFinder1.3, Interface1.0 [bne]
    :Imports.    ModuleInfos1.2, ErrorReq1.2, MemSystem1.3 [bne]
    :History.    V1.0b [bne] 4.Feb.1989
    :History.    V1.1a [bne] 9.Feb.1989 (+ ModQueue)
    :History.    V1.2b [bne] 17.Feb.1989 (bugs fixed, + Interface)
    :Remark.      V1.2 first release, alpha test version
    :History.    V1.3b [bne] 22.Feb.1989 (InitGlobal, cosmetics)
    :Update.     4.Mar.1989 [bne] adapted MemSystem V1.3
    :Remark.      V1.3 second release, beta test version
    :History.    V1.4 [bne] 4.Apr.1989 (gamma release, shareware)
    :History.    V1.5 [bne] 8.Apr.1989 (+"a"-Option)

**********************************************************************)

MODULE Make;

FROM Arts       IMPORT maxModName, ModName, ModuleId, wbStarted, ModType,
                ModKeys;
FROM Dos        IMPORT FileLockPtr, CurrentDir, Lock, UnLock, sharedLock,
                Date;
FROM ErrorReq   IMPORT Assert;
FROM Environment IMPORT OptionFlags, OptionFlagSet, InitInOut, CloseInOut,
                WriteString, WriteLn, GetArguments;
FROM Interface  IMPORT InitCompilerArgList, AddCompilerArg, StartCompiler,
                StartLinker, WaitUntilCompleted;
FROM MemSystem  IMPORT NoCareAllocate, Deallocate;
FROM ModuleInfos IMPORT ReadDefInfo, ReadModInfo, ModuleInfo,
                ModuleInfoPtr, MatchKeys, ImportInfo, ImportInfoPtr,
                ReadMainInfo, InfoFlags, InfoFlagSet, GetDate,
                ResetLastFileDate, LastFileDate, EarlierDate;
FROM PathFinder IMPORT ExtendFileName, ModNamePtr, ProjectLock,
                MainProject, DirLock, PathFlags, InitGlobal, BinDir,
                GoToProject;
FROM SYSTEM     IMPORT ADR, ADDRESS;
IMPORT AVL,ASCII,Queue;

CONST
  Copyright="M2Make (c) 1989 by Nicolas Benezan";

  (* Error messages: *)
  QueueInitFail   = "Queue.Init() failed";
  AVLInitFail     = "AVL.Init(ModuleHeap) failed";
  AVLWriteFail    = "AVL.Write() failed";
  TwiceSameName   = "twice the same module name";
  EnqueueFail     = "Enqueue() failed";
  ModuleNotInAVL  = "Module not in the AVL.tree";

TYPE
  ModStatus=(noDef(*main module*),newDef,oldDef);
  ImplementInfo=RECORD
    name:ModName;
    status:ModStatus;
  END;

VAR
  MainModule:ModName;
  MainInfo:ModuleInfoPtr;
  Options:OptionFlagSet;
  AnyChanges:BOOLEAN;
  Import:ImportInfoPtr;
  ModuleHeap:AVL.tree;
  ModQueue:Queue.queue;
  Status:ModStatus;
  Implementation:ModName;
  ExeFileDate:Date;

(**)
(* key compare procedure for AVL-tree handling                          *)
(**)
PROCEDURE CompareNames(Name1,Name2:ADDRESS):INTEGER;
  VAR
    String1,String2:POINTER TO CHAR;
  BEGIN
    String1:=Name1;
    String2:=Name2;
    WHILE (String1^=String2^) AND (String1^#ASCII.nul) DO
      INC(String1);
      INC(String2);
    END;
    RETURN ORD(String1^)-ORD(String2^);
  END CompareNames;

(**)
(* Add an entry to the compile list                                     *)
(**)
PROCEDURE Compile(Project:ProjectLock;Name:ModName;Type:PathFlags);
  BEGIN
    ExtendFileName(Name,Type);
    WriteString(" * ");WriteString(Name);WriteString(" to be compiled");
    WriteLn;
    AddCompilerArg(Name,DirLock(Project,Type));
    AnyChanges:=TRUE;
  END Compile;

(**)
(* Start the linker                                                     *)
(**)
PROCEDURE Link(Name:ModName);
  BEGIN
    ExtendFileName(Name,obj);
    WriteString("linking ");WriteString(Name);
    WriteLn;
    StartLinker(Name,DirLock(MainProject(),obj));
    WaitUntilCompleted;
  END Link;

(**)
(* install memory allocation procedures                                 *)
(**)
PROCEDURE InitMemSystem;
  BEGIN
    AVL.AvlAllocProc:=NoCareAllocate;
    AVL.AvlDeallocProc:=Deallocate;
    Queue.QueueAllocProc:=NoCareAllocate;
    Queue.QueueDeallocProc:=Deallocate;
  END InitMemSystem;

(**)
(* Read module information from the AVL.tree                            *)
(**)
PROCEDURE GetProjectInfo(Name:ModNamePtr;VAR ModInfo:ModuleInfoPtr):BOOLEAN;
  BEGIN
    RETURN AVL.Get(ModuleHeap,Name^,ModInfo);
  END GetProjectInfo;

(**)
(* Write module information to the AVL.tree                             *)
(**)
PROCEDURE AddProjectInfo(Name:ModNamePtr;ModInfo:ModuleInfoPtr);
  VAR
    Over:BOOLEAN;
  BEGIN
    WriteString(" - ");WriteString(Name^);
    WriteLn;
    Assert(AVL.Write(ModuleHeap,Name^,ModInfo^,Over),ADR(AVLWriteFail));
    Deallocate(ModInfo); (* ModInfo^ was copied into AVL.tree *)
    Assert(NOT Over,ADR(TwiceSameName));
  END AddProjectInfo;

(**)
(* Enqueue iplementation module for later processing                    *)
(**)
PROCEDURE EnqueueMod(Name:ModNamePtr;Status:ModStatus);
  VAR
    Module:ImplementInfo;
  BEGIN
    Module.name:=Name^;
    Module.status:=Status;
    Assert(Queue.Write(ModQueue,Module),ADR(EnqueueFail));
  END EnqueueMod;

(**)
(* Dequeue implementation module                                        *)
(**)
PROCEDURE DequeueMod(VAR Name:ModName;VAR Status:ModStatus);
  VAR
    Module:ImplementInfo;
  BEGIN
    Queue.Read(ModQueue,Module);
    Name:=Module.name;
    Status:=Module.status;
  END DequeueMod;

(**)
(* Examine DEFINITION MODULE - Compile() if necessary                   *)
(**)
PROCEDURE ExamineDef(Project:ProjectLock; (* info for PathFinder *)
                     Module:ModNamePtr;
                     ExpectedKey:ModKeys):BOOLEAN (* new version ? *);
  VAR
    NewVersion:BOOLEAN;
    ModInfo:ModuleInfoPtr;
    ImportList:Queue.queue; (* queue instead of list for simplification *)
  (*Import:ImportInfoPtr is global for efficiency reasons *)
  BEGIN
    IF GetProjectInfo(Module,ModInfo) THEN       (* if already examined *)
      WITH ModInfo^ DO
        RETURN compiled OR NOT MatchKeys(ExpectedKey,key);
      END;
    ELSE
      Assert(Queue.Init(ImportList),ADR(QueueInitFail));
      ReadDefInfo(Project,Module,ImportList,ModInfo);
      NewVersion:=FALSE;
      WHILE NOT Queue.Empty(ImportList) DO (* before compiling this Module we *)
        Queue.Get(ImportList,Import);     (* Examine() all imported Modules *)
        NewVersion:=(ExamineDef(ModInfo^.project,ADR(Import^.name),Import^.key)
                     OR NewVersion);
        Queue.Remove(ImportList);
      END;
      Queue.Discard(ImportList);
      NewVersion:=NewVersion OR
                  ((all IN Options) AND (defExist IN ModInfo^.flags))
                  OR NOT (symValid IN ModInfo^.flags);
      IF NewVersion THEN
        Compile(ModInfo^.project,Module^,def);
        ModInfo^.compiled:=TRUE;
        Status:=newDef;
      ELSE
        Status:=oldDef;
      END;
      IF ModInfo^.type=mod THEN
        EnqueueMod(Module,Status); (* Implementation Module *)
      END;
      NewVersion:=NewVersion OR NOT MatchKeys(ExpectedKey,ModInfo^.key);
      AddProjectInfo(Module,ModInfo);         (* prevent Module from being compiled twice *)
        (* ModInfo is invalid after AddProjectInfo() !!! *)
      RETURN NewVersion;
    END;
  END ExamineDef;

(**)
(* Examine IMPLEMENTATION MODULE - Compile() if neccessary              *)
(**)
PROCEDURE ExamineMod(Module:ModNamePtr;
                     Status:ModStatus);
  VAR
    NewMod:BOOLEAN;
    ImportList:Queue.queue; (* queue instead of list for simplification *)
    ModInfo:ModuleInfoPtr;
    ObjKey:ModKeys;
  (*Import:ImportInfoPtr is global for efficiency reasons *)
  BEGIN
    NewMod:=FALSE;
    Assert(GetProjectInfo(Module,ModInfo),ADR(ModuleNotInAVL));
    Assert(Queue.Init(ImportList),ADR(QueueInitFail));
    ReadModInfo(Module,ImportList,ModInfo,ObjKey);
    WHILE NOT Queue.Empty(ImportList) DO
      Queue.Get(ImportList,Import);
      NewMod:=(ExamineDef(ModInfo^.project,ADR(Import^.name),Import^.key) OR NewMod);
      Queue.Remove(ImportList);
    END;
    Queue.Discard(ImportList);
    IF NOT NewMod AND (Status#newDef) AND (objValid IN ModInfo^.flags)
       AND NOT ((all IN Options) AND (modExist IN ModInfo^.flags))
       AND ((Status=noDef) OR MatchKeys(ModInfo^.key,ObjKey)) THEN
      RETURN
    END;
    Compile(ModInfo^.project,Module^,imp);
  END ExamineMod;

(**)
(* Get date of executable (if exists) and reset LatestFileDate          *)
(**)
PROCEDURE InitDates;
  VAR
    BinDirLock,OldDir,ExeLock:FileLockPtr;
  BEGIN
    GoToProject(MainProject(),proj);
    BinDirLock:=Lock(ADR(BinDir),sharedLock);
    IF BinDirLock#NIL THEN
      OldDir:=CurrentDir(BinDirLock);
    END;
    ExeLock:=Lock(ADR(MainModule),sharedLock);
    IF ExeLock#NIL THEN
      GetDate(ExeLock,ExeFileDate);
      ResetLastFileDate;
      UnLock(ExeLock);
    ELSE
      ResetLastFileDate;
      ExeFileDate:=LastFileDate; (* null *)
    END;
    IF BinDirLock#NIL THEN
      BinDirLock:=CurrentDir(OldDir);
      UnLock(BinDirLock);
    END;
  END InitDates;

(**)
(* Make Main                                                            *)
(**)
BEGIN
  InitMemSystem;
  GetArguments(MainModule,Options);
  InitInOut;
  InitGlobal;
  WriteString("Amiga Modula-2 Make, gamma release, 8.Apr.89,  AMOK [bne]");
  WriteLn;
  WriteString("This is SHAREWARE, not in the Public Domain !");
  WriteLn;
  WriteString("making ");WriteString(MainModule);
  WriteLn;
  AnyChanges:=FALSE;
  Assert(AVL.Init(ModuleHeap,"CheckedModules.tree",
         SIZE(ModName),SIZE(ModuleInfo),CompareNames),ADR(AVLInitFail));
  Assert(Queue.Init(ModQueue),ADR(QueueInitFail));
  InitCompilerArgList;
  ReadMainInfo(MainProject(),ADR(MainModule),MainInfo);
  AddProjectInfo(ADR(MainModule),MainInfo);
  InitDates;
  ExamineMod(ADR(MainModule),noDef);
  WHILE NOT Queue.Empty(ModQueue) DO
    DequeueMod(Implementation,Status);
    ExamineMod(ADR(Implementation),Status);
  END;
  Queue.Discard(ModQueue);
  AVL.Discard(ModuleHeap);
  IF AnyChanges THEN
    StartCompiler;
    WaitUntilCompleted;
  END;
  IF AnyChanges OR EarlierDate(ExeFileDate,LastFileDate) THEN
    Link(MainModule);
  END;
  CloseInOut;
END Make.

