(**********************************************************************

    :Program.    PathFinder.mod
    :Contents.   m2Amiga project path handler for Make [bne]
    :Author.     Nicolas Benezan [bne]
    :Address.    Postwiesenstr. 2, D7000 Stuttgart 60
    :Phone.      711/333679
    :Copyright.  Shareware, see Doc-File
    :Language.   Modula-2
    :Translator. M2Amiga A+L V3.2d
    :Imports.    Environment1.2, MemSystem1.3, ErrorReq [bne]
    :Imports.    AVL1.2, List1.1 [mif]
    :History.    V1.0 [bne] 7.Feb.1989
    :History.    V1.1 [bne] 10.Feb.1989 (+ FindLocal)
    :History.    V1.2 [bne] 14.Feb.1989 (bugs fixed)
    :History.    V1.3 [bne] 15.Feb.1989 (+ DirLock)
    :History.    V1.4d [bne] 22.Feb.1989 (+ Assert(globalPathLoaded))
    :Update.     4.Mar.1989 [bne] adapted MemSystem V1.3
    :History.    V1.5 [bne] 4.Apr.1989 (+ BinDir, cosmetics)

**********************************************************************)

IMPLEMENTATION MODULE PathFinder;

FROM Arts       IMPORT ModName,TermProcedure;
FROM Dos        IMPORT FileLockPtr,FileInfoBlock,FileInfoBlockPtr,
                ParentDir,Lock,DupLock,UnLock,sharedLock,Examine,
                FileHandlePtr,Open,Close,Read,oldFile;
FROM MemSystem  IMPORT NoCareAllocate,Deallocate;
FROM Environment IMPORT SystemDir,MainDir,CurrentDir,SetDir;
FROM ErrorReq   IMPORT Assert;
FROM Strings    IMPORT Insert,Compare;
FROM Str        IMPORT Concat,Length;
FROM SYSTEM     IMPORT ADR,ADDRESS,SETREG,BPTR;
IMPORT AVL,List,Dos,ASCII;

        (* Error messages *)
CONST
  ProjNotFound    = "ProjectLock not found";
  AddProjFail     = "AddProject() failed";
  ListInitFail    = "List.Init(lockList) failed";
  NoPathFile      = "unlikely to be a path file";
  PathNotOpened   = "can't Open(PathFile)";
  PathNotFound    = "path not found";
  InitAVLFail     = "Init(AVL.tree) failed";
  NoGlobalPath    = "global path not found";

TYPE
  ProjectLock=FileLockPtr;
  PathListStatus=(unknown,loaded,examined);
  CharPtr=POINTER TO CHAR;
  PathList=RECORD
    lockList:List.list;
    bufferPos:CharPtr;
    bufferEnd:CharPtr;
    status:PathListStatus;
  END;
  ProjectInfo=RECORD
    flags:PathFlagSet;
    subDir:ARRAY [sym..imp] OF FileLockPtr;  (* sym,obj,txt *)
    localPath:PathList;
  END;
  ProjectInfoPtr=POINTER TO ProjectInfo;

VAR
  LockTree:AVL.tree;
  GlobalPath:PathList;
  CurrentProject,MainLock:ProjectLock;
  CurrentInfo:ProjectInfoPtr;
  OriginDir:FileLockPtr;

(**)
(* Compare procedure for AVL.tree                                       *)
(**)
PROCEDURE ComparePointers(Adr1,Adr2:ADDRESS):INTEGER;
  VAR
    Ptr1,Ptr2:POINTER TO LONGINT;
  BEGIN
    Ptr1:=Adr1;
    Ptr2:=Adr2;
    IF Ptr1^<Ptr2^ THEN
      RETURN -1;
    ELSIF Ptr1^=Ptr2^ THEN
      RETURN 0;
    ELSE
      RETURN 1;
    END;
  END ComparePointers;

(**)
(* Full path name of the current directory                              *)
(**)
PROCEDURE GetCurrentPath(VAR Path:ARRAY OF CHAR);
  VAR
    Info:FileInfoBlockPtr;
    Pos:INTEGER;
    Parent,Dir:FileLockPtr;
  BEGIN
    Path:="";
    Pos:=0;
    NoCareAllocate(Info,SIZE(Info^));
    Dir:=DupLock(CurrentDir());
    IF Info#NIL THEN
      WHILE Dir#NIL DO
        IF Examine(Dir,Info) AND (Info^.dirEntryType>0) THEN
          Insert(Path,0,"/");
          Insert(Path,0,Info^.fileName);
          Pos:=Length(Info^.fileName);
        END;
        Parent:=ParentDir(Dir);
        UnLock(Dir);
        Dir:=Parent;
      END;
      IF Pos#0 THEN
        Path[Pos]:=":";
      END;
      Deallocate(Info);
    END;
  END GetCurrentPath;

PROCEDURE DirLock(Project:ProjectLock;FileType:PathFlags):FileLockPtr;
  BEGIN
    IF Project#CurrentProject THEN
      CurrentProject:=Project;
      Assert(AVL.Get(LockTree,CurrentProject,CurrentInfo),ADR(ProjNotFound));
    END;
    WITH CurrentInfo^ DO
      IF FileType IN flags THEN
        RETURN subDir[FileType];
      ELSE
        RETURN BPTR(Project);
      END;
    END;
  END DirLock;

(**)
(* Set current dir to a project (and probably sym/obj/txt subdirs)      *)
(**)
PROCEDURE GoToProject(Project:ProjectLock;FileType:PathFlags);
  BEGIN
    SetDir(DirLock(Project,FileType));
  END GoToProject;

(**)
(* Examine a project dir and add it to the AVL.tree                     *)
(**)
PROCEDURE AddProject(NewProjectDir:FileLockPtr):ProjectLock;
  VAR
    Info:ProjectInfo;
    NewProject:ProjectLock;
    SubDir:FileLockPtr;
    Over:BOOLEAN; (* Dummy *)
  BEGIN
    NewProject:=BPTR(NewProjectDir);
    SetDir(NewProjectDir);
    WITH Info DO
      SubDir:=Lock(ADR(DefDir),sharedLock); (* "txt" *)
      IF SubDir#NIL THEN
        subDir[def]:=SubDir;
        subDir[imp]:=SubDir;
        flags:=PathFlagSet{def,imp};
      ELSE
        flags:=PathFlagSet{};
      END;
      SubDir:=Lock(ADR(SymDir),sharedLock); (* "sym" *)
      IF SubDir#NIL THEN
        subDir[sym]:=SubDir;
        INCL(flags,sym);
      END;
      SubDir:=Lock(ADR(ObjDir),sharedLock); (* "obj" *)
      IF SubDir#NIL THEN
        subDir[obj]:=SubDir;
        INCL(flags,obj);
      END;
      localPath.status:=unknown; (* path file has not been examined yet *)
    END; (* with Info *)
    Assert(AVL.Write(LockTree,NewProject,Info,Over),ADR(AddProjFail));
    RETURN NewProject;
  END AddProject;

(**)
(* Load a local path file (if found)                                    *)
(**)
PROCEDURE LoadPathFile(VAR list:PathList);
  VAR
    PathFileLock:FileLockPtr;
    PathFile:FileHandlePtr;
    InfoBlock:FileInfoBlockPtr;
  BEGIN
    NoCareAllocate(InfoBlock,SIZE(FileInfoBlock));
    WITH list DO
      status:=examined;
      Assert(List.Init(lockList),ADR(ListInitFail));
      PathFileLock:=Lock(ADR(PathFileName),sharedLock); (* do not terminate before UnLock() ! *)
      IF PathFileLock#NIL THEN
        IF Examine(PathFileLock,InfoBlock) THEN
          UnLock(PathFileLock);
          WITH InfoBlock^ DO
            Assert((size>1) AND (size<2000),ADR(NoPathFile));
            NoCareAllocate(bufferPos,size);
            PathFile:=Open(ADR(PathFileName),oldFile);
            Assert(PathFile#NIL,ADR(PathNotOpened));
            IF Read(PathFile,bufferPos,size)=size THEN
              status:=loaded;
              bufferEnd:=CharPtr(LONGINT(bufferPos)+size);
            END;
            Close(PathFile);
          END; (* with Info Block *)
        ELSE
          UnLock(PathFileLock);
        END; (* if Examine() *)
      END; (* if PathFileLock#NIL *)
    END; (* with list *)
    Deallocate(InfoBlock);
  END LoadPathFile;

(**)
(* scan path file, mark end of line with ASCII.nul                      *)
(**)
PROCEDURE ScanBuffer(VAR Pos:CharPtr;End:CharPtr):ADDRESS;
  VAR
    StartPos:CharPtr;
  BEGIN
    StartPos:=Pos;
    DEC(End);
    WHILE (Pos^>=" ") AND (LONGINT(Pos)<LONGINT(End)) DO
      INC(Pos);       (* skip pathname *)
    END;
    INC(End);
    Pos^:=ASCII.nul;
    WHILE (Pos^<" ") AND (LONGINT(Pos)<LONGINT(End)) DO
      INC(Pos);       (* skip lf, cr *)
    END;
    RETURN ADDRESS(StartPos);
  END ScanBuffer;

(**)
(* Append an extension to a file name                                   *)
(**)
PROCEDURE ExtendFileName(VAR Name:ARRAY OF CHAR;Type:PathFlags);
  VAR
    Extens:ARRAY [0..4] OF CHAR;
  BEGIN
    CASE Type OF
      |def: Extens:=DefExtens;
      |imp: Extens:=ModExtens;
      |sym: Extens:=SymExtens;
      |obj: Extens:=ObjExtens;
    ELSE
      RETURN (* FileType=proj *)
    END;
    Concat(Name,Extens);
  END ExtendFileName;

(**)
(* Search a module tracing the paths (first local then global)          *)
(**)
PROCEDURE FindProject(VAR CurrentProj:ProjectLock;
                          ModuleName:ModName;
                          FileType:PathFlags):FileLockPtr;
VAR
  File:FileLockPtr;
  Info:ProjectInfoPtr;

  PROCEDURE SearchPath(VAR list:PathList);
    VAR
      NewLock:FileLockPtr;
      NewPath:POINTER TO ARRAY [0..80] OF CHAR;
    BEGIN
      WITH list DO
        List.Reset(lockList);
        LOOP
          IF List.End(lockList) THEN
            IF status=loaded THEN
              NewPath:=ScanBuffer(bufferPos,bufferEnd);
              NewLock:=Lock(NewPath,sharedLock);
              Assert(NewLock#NIL,ADR(PathNotFound));
              CurrentProj:=AddProject(NewLock);
              IF LONGINT(bufferPos)>=LONGINT(bufferEnd) THEN
                status:=examined;
              END;
              List.Write(lockList,CurrentProj);
            ELSE
              EXIT
            END;
          ELSE
            List.Read(lockList,CurrentProj);
          END;
          GoToProject(CurrentProj,FileType);
          File:=Lock(ADR(ModuleName),sharedLock);
          IF File#NIL THEN
            EXIT
          END;
        END;
      END;
    END SearchPath;

  BEGIN
    ExtendFileName(ModuleName,FileType);
    GoToProject(CurrentProj,FileType);    (* set CurrentDir *)
    File:=Lock(ADR(ModuleName),sharedLock);
    IF File=NIL THEN
      Info:=CurrentInfo;
      IF Info^.localPath.status=unknown THEN
        GoToProject(CurrentProj,proj);
        LoadPathFile(Info^.localPath);
      END;
      SearchPath(Info^.localPath);
      IF File=NIL THEN
        SearchPath(GlobalPath);
      END;
    END;
    RETURN File;
  END FindProject;

(**)
(* Find a file in a project drawer or sym/obj/txt subdirs               *)
(**)
PROCEDURE FindLocal(CurrentProj:ProjectLock;
                    ModuleName:ModName;
                    FileType:PathFlags):FileLockPtr;
  BEGIN
    GoToProject(CurrentProj,FileType);
    ExtendFileName(ModuleName,FileType);
    RETURN Lock(ADR(ModuleName),sharedLock);
  END FindLocal;

(**)
(* Root of path searching                                               *)
(**)
PROCEDURE MainProject():ProjectLock;
  BEGIN
    RETURN MainLock;
  END MainProject;

(**)
(* Cleanup the AVL.tree                                                 *)
(**)
PROCEDURE FreeLocks(LockPtrAdr,InfoAdr:ADDRESS);
  VAR
    LockPtr:POINTER TO FileLockPtr;
    SubFlag:PathFlags;
    InfoPtr:ProjectInfoPtr;
  BEGIN
    LockPtr:=LockPtrAdr;
    InfoPtr:=InfoAdr;
    WITH InfoPtr^ DO
      FOR SubFlag:=sym TO def DO (* sym, obj, txt *)
        IF SubFlag IN flags THEN
          UnLock(subDir[SubFlag]);
        END;
      END;
      IF localPath.status#unknown THEN
        List.Discard(localPath.lockList);
      END;
    END;
    UnLock(LockPtr^); (* proj *)
  END FreeLocks;

(**)
(* Termination procedure                                                *)
(**)
PROCEDURE CleanUp;
  BEGIN
    SetDir(OriginDir); (* UnLock(CurrentDir) would be deadly *)
    AVL.Do(LockTree,FreeLocks,AVL.post);
    AVL.Discard(LockTree);
  END CleanUp;

(**)
(* Examine the main project, load the global path                       *)
(**)
PROCEDURE InitGlobal;
  VAR
    OldDir:FileLockPtr;
    MainPath:ARRAY [0..80] OF CHAR;
    Len:INTEGER;
  BEGIN
    Assert(AVL.Init(LockTree,"PathFinder.tree",SIZE(ProjectLock),SIZE(ProjectInfo),
                ComparePointers),ADR(InitAVLFail));
    TermProcedure(CleanUp);
    OldDir:=Dos.CurrentDir(MainDir);
    GetCurrentPath(MainPath);
    Len:=Length(MainPath);
    IF (Len>4) AND (Compare(MainPath,Len-4,3,ModDir,FALSE)=0) THEN
      MainLock:=AddProject(ParentDir(MainDir)); (* project/txt/MainModule *)
    ELSE
      MainLock:=AddProject(DupLock(MainDir));   (* project/MainModule *)
    END;
    SetDir(SystemDir);
    LoadPathFile(GlobalPath);
    Assert(GlobalPath.status=loaded,ADR(NoGlobalPath));
    SetDir(OldDir);
  END InitGlobal;

BEGIN
  MainLock:=NIL;
  CurrentProject:=NIL;
END PathFinder.

